/*
 * $Id: icontrolprefs.c,v 1.6 2007/11/15 11:18:53 piru Exp $
 */
#define AROS_ALMOST_COMPATIBLE
#include "global.h"
#include <proto/iffparse.h>
#include <aros/asmcall.h>
#include <devices/rawkeycodes.h>
#include <libraries/commodities.h>

#define DEBUG 0
#include <aros/debug.h>

#define DEBUG_OLDFORMAT(x) ;

struct Screen *FindWBScreen(void);
void ReOpenWBScreen(struct Screen *MyScreen);

struct IIControlPrefs
{
	UWORD ic_TimeOut;
	WORD  ic_MetaDrag;
	ULONG ic_Flags;
	UBYTE ic_WBtoFront;
	UBYTE ic_FrontToBack;
	UBYTE ic_ReqTrue;
	UBYTE ic_ReqFalse;
};

AROS_UFH1(LONG, IControlPrefs_Handler,
		  AROS_UFHA(struct IFFHandle *, iff, A2))
{
	struct IControlPrefs iprefs;
	struct IIControlPrefs oldiprefs;

	ReadChunkBytes(iff, &iprefs, sizeof(iprefs));
	SetIPrefs(&oldiprefs,sizeof (oldiprefs),IP_OLDICONTROL);
	SetIPrefs(&iprefs.ic_TimeOut, sizeof(iprefs) - 16, IP_ICONTROL);
	if ((oldiprefs.ic_Flags & ((ICF_NOWBTITLEBAR|ICF_DISAPPEARINGTITLEBAR))) !=
		(iprefs.ic_Flags & (ICF_NOWBTITLEBAR|ICF_DISAPPEARINGTITLEBAR)))
	{
		struct Screen *MyScreen;

		D(bug("FontPrefs_Handler: SetIPrefs\n"));

		MyScreen = FindWBScreen();

		if (MyScreen) ReOpenWBScreen(MyScreen);

	}

	return 0;

	AROS_LIBFUNC_EXIT
}

AROS_UFH1(LONG, IControlExtensions_Handler,
		  AROS_UFHA(struct IFFHandle *, iff, A2))
{
	struct IControlExtensions iprefs;

	ReadChunkBytes(iff, &iprefs, sizeof(iprefs));
	SetIPrefs(&iprefs, sizeof(iprefs), IP_IEXTENSIONS);

	return 0;

	AROS_LIBFUNC_EXIT
}


AROS_UFH1(LONG, IControlHotkeys_Handler, AROS_UFHA(struct IFFHandle *, iff, A2))
{
	ULONG *mem = 0,len = 0;
	struct ContextNode *cn;

	cn = CurrentChunk(iff);

	if (cn)
	{
		len = cn->cn_Size;
		mem = AllocMem(len,MEMF_ANY);

		if (mem)
		{
			ReadChunkBytes(iff, mem, len);

			SetIPrefs(mem,len, IP_IACTIONS);

			FreeMem(mem,len);
		}
	}

	return 0;

	AROS_LIBFUNC_EXIT
}

/* loader for old prefs format */
struct OldIAction
{
	UWORD ia_Key;
	UWORD ia_Qualifier;
	ULONG ia_Special;
};

#define NUM_OLDIACTIONS 30
#define ACTIVATIONFLAG 0x4

struct OldStruct
{
	ULONG nothing[4];
	struct OldIAction oldiaction[NUM_OLDIACTIONS];
};


AROS_UFH1(LONG, IControlActions_Handler, AROS_UFHA(struct IFFHandle *, iff, A2))
{
	struct OldStruct iactions;
	ULONG numhotkeys = 0,no;

	ReadChunkBytes(iff, &iactions, sizeof(iactions));

	for (no = 0; no < NUM_OLDIACTIONS; no ++)
	{
		if (iactions.oldiaction[no].ia_Special & ACTIVATIONFLAG)
		numhotkeys ++;
	}

	DEBUG_OLDFORMAT(dprintf("IControl: found %ld hotkeys\n",numhotkeys))

	if (numhotkeys)
	{
		struct IAction *ia = AllocVec((sizeof (struct IAction)) * numhotkeys,MEMF_ANY);

		if (ia)
		{
			ULONG currenthot = 0;

			DEBUG_OLDFORMAT(dprintf("IControl: mem allocated\n"))

			for (no = 0; no < NUM_OLDIACTIONS; no ++)
			{
				if (iactions.oldiaction[no].ia_Special & ACTIVATIONFLAG)
				{
					ia[currenthot].ia_Action = no;
					ia[currenthot].ia_Key = iactions.oldiaction[no].ia_Key;
					ia[currenthot].ia_Qualifier = iactions.oldiaction[no].ia_Qualifier;
					ia[currenthot].ia_Special = iactions.oldiaction[no].ia_Special & ~ACTIVATIONFLAG;
					currenthot ++;
				}

			}

			DEBUG_OLDFORMAT(dprintf("IControl: SetIPrefs (%lx, size = %ld, hotkeys %ld)\n",ia,(sizeof (struct IAction)) * numhotkeys,currenthot))
			SetIPrefs(ia,(sizeof (struct IAction)) * numhotkeys, IP_IACTIONS);
			
			FreeVec(ia);
		}
	}


	return 0;

	AROS_LIBFUNC_EXIT
}

char *int_skipspaces(char *string)
{
	while ((string[0] == 0x9) || (string[0] == 0x20))
	{
		string++;
	}

	return string;
}

WORD int_wordlen(char *string)
{
	WORD len = 0;

	for(;;)
	{
		if (string[len] == 0x00) return len;
		if (string[len] == 0x09) return len; //tab
		if (string[len] == 0x20) return len; //space
		if (string[len] == 0x0A) return len; //\n
		if (string[len] == 0x0D) return len; //\r
		if (string[len] == 0x2C) return len; //,
		if (string[len] == 0x3B) return len; //;
		len++;
	}
	return len;
}

char *int_skipword(char *string)
{
	for (;;)
	{
		if (string[0] == 0x00) return string;
		if (string[0] == 0x09) return string;
		if (string[0] == 0x20) return string;
		if (string[0] == 0x0A) return string;
		if (string[0] == 0x0D) return string;
		if (string[0] == 0x2C) return string;
		if (string[0] == 0x3B) return string;
		string ++;
	}
}

void int_getword(char *from,char *to)
{
	WORD len = int_wordlen(from);

	if (len > 0)
	{
		memcpy(to,from,len);
		to[len] = 0;
	} else {
		to[0] = 0;
	}
}

char *int_nexthotkey(char *string)
{
	for (;;)
	{
		if (string[0] == 0x0) return string;
		if (string[0] == 0x2c) return string; // ,
		string++;
	}
}

#define mouse_leftpress 0x68
#define mouse_middlepress 0x6a
#define mouse_rightpress 0x69

static void translate(char *src,struct IAction *action)
{
	struct KeyIdent { const char *name; int code; };

	static const struct KeyIdent qualis[] =
	{
		{"shift",IXSYM_SHIFTMASK},
		{"caps",IXSYM_CAPSMASK},
		{"alt",IXSYM_ALTMASK},
		{"lshift",IEQUALIFIER_LSHIFT},
		{"rshift",IEQUALIFIER_RSHIFT},
		{"capslock",IEQUALIFIER_CAPSLOCK},
		{"control",IEQUALIFIER_CONTROL},
		{"lalt",IEQUALIFIER_LALT},
		{"ralt",IEQUALIFIER_RALT},
		{"lcommand",IEQUALIFIER_LCOMMAND},
		{"rcommand",IEQUALIFIER_RCOMMAND},
	};

	static const struct KeyIdent keys[] =
	{
		{"left",0x4f},
		{"right",0x4e},
		{"up",0x4c},
		{"down",0x4d},
		{"backspace",0x41},
		{"delete",0x46},
		{"comma",0x38},
		{"break",0x6e},
		{"end",0x71},
		{"enter",0x43},
		{"esc",0x45},
		{"f1",0x50},
		{"f2",0x51},
		{"f3",0x52},
		{"f4",0x53},
		{"f5",0x54},
		{"f6",0x55},
		{"f7",0x56},
		{"f8",0x57},
		{"f9",0x58},
		{"f10",0x59},
		{"f11",0x4b},
		{"f12",0x6f},
		{"help",0x5f},
		{"home",0x70},
		{"insert",0x47},
		{"numlock",0x6d},
		{"page_down",0x49},
		{"page_up",0x48},
		{"pause",0x6e},
		{"prtscr",0x6c},
		{"return",0x44},
		{"scrlock",0x6b},
		{"space",0x40},
		{"tab",0x42},
#if 0 /* Don't frigging convert these to keypad rawkeys... */
		{"(",0x5a},
		{")",0x5b},
		{"*",0x5d},
		{"+",0x5e},
		{"-",0x4a},
		{".",0x3c},
		{"/",0x5c},
		{"0",0x0f},
		{"1",0x1d},
		{"2",0x1e},
		{"3",0x1f},
		{"4",0x2d},
		{"5",0x2e},
		{"6",0x2f},
		{"7",0x3d},
		{"8",0x3e},
		{"9",0x3f},
#endif
		{"wheel_up",0x7a},
		{"wheel_down",0x7b},
		{"wheel_left",0x7c},
		{"wheel_right",0x7d},
		{"media_stop",0x72},
		{"media_play",0x73},
		{"media_prev",0x74},
		{"media_next",0x75},
		{"media_rewind",0x76},
		{"media_forward",0x77},
		{"mouse_leftpress",mouse_leftpress},
		{"mouse_middlepress",mouse_middlepress},
		{"mouse_rightpress",mouse_rightpress},
		{"mouse_fourthpress",0x7e},
	};

	while (src && *src)
	{
		char word[256];
		BOOL found = FALSE;
		int it;

		src = int_skipspaces(src);

		if (!*src || *src == ',') return; // end processing

		int_getword(src,word);

		for (it = 0; it < sizeof (qualis)/sizeof (struct KeyIdent); it++)
		{
			if (!strcmp(word,qualis[it].name))
			{
				action->ia_Qualifier |= qualis[it].code;
				found = TRUE;
				break;
			}
		}

		if (!found)
		for (it = 0; it < sizeof (keys)/sizeof (struct KeyIdent); it++)
		{
			if (!strcmp(word,keys[it].name))
			{
				action->ia_Key = keys[it].code;
				found = TRUE;
				break;
			}
		}

		if (!found)
		{
			if (!strcmp("double",word)) {action->ia_Special = IASP_DOUBLECLICK; found = TRUE;}
			if (!strcmp("triple",word)) {action->ia_Special = IASP_TRIPLECLICK; found = TRUE;}
		}

		if (!found)
		{
			// mapansi as the last resort :)
			UBYTE ansibuffer[2];

			if (MapANSI(word,strlen(word),ansibuffer,1,NULL) == 1)
			{
				action->ia_Key = ansibuffer[0];
			}
		}

		src = int_skipword(src);
	}
}

static const UBYTE options[] =
	"Flags/N/K,"
	"Hotkey/N/K,"
	"Trigger/K,"
	"End/S,"
	"/F";

struct Args
{
	ULONG *Flags;
	ULONG *Hotkey;
	STRPTR Trigger;
	ULONG  EndOfFile;
	STRPTR FutureArgs;
};

void LoadIControl(char *filename)
{
	ULONG numkeys = 0; ULONG keysleft = 255;
	const ULONG alen = sizeof (struct IAction) * keysleft + 1;
	struct IAction *actions = AllocMem(alen, MEMF_CLEAR|MEMF_PUBLIC);
	BPTR file;

	if (!actions) return;
	
	if ((file = Open(filename,MODE_OLDFILE)))
	{
		BPTR OldFile;
		struct RDArgs rda;
		struct RDArgs *rda1;
		struct Args args;

		for (;;)
		{
			memset(&rda,0,sizeof(struct RDArgs));

			OldFile = Input();
			SelectInput(file);
			rda.RDA_Flags = RDAF_NOPROMPT | RDAF_STDIN;

			memset(&args,0,sizeof(args));

			if ((rda1 = ReadArgs(options, (LONG *)&args, &rda)))
			{
				if (args.Flags)
				{
					struct IIControlPrefs oldiprefs;
					ULONG oldflags;

					SetIPrefs(&oldiprefs,sizeof (oldiprefs),IP_OLDICONTROL);
					
					oldflags = oldiprefs.ic_Flags;
					oldiprefs.ic_Flags = *args.Flags;

					SetIPrefs(&oldiprefs.ic_TimeOut, sizeof(oldiprefs), IP_ICONTROL);

					if ((oldflags & ((ICF_NOWBTITLEBAR|ICF_DISAPPEARINGTITLEBAR))) !=
						(oldiprefs.ic_Flags & (ICF_NOWBTITLEBAR|ICF_DISAPPEARINGTITLEBAR)))
					{
						struct Screen *MyScreen;

						MyScreen = FindWBScreen();

						if (MyScreen) ReOpenWBScreen(MyScreen);
					}
				}
				else if	(args.Hotkey && args.Trigger)
				{
					char *keystring = args.Trigger;

					while (keysleft > 0)
					{
						actions[numkeys].ia_Action = *args.Hotkey;

						// parse the hotkey definition now!
						translate(keystring,&actions[numkeys]);

						keysleft --;
						numkeys++;

						keystring = int_nexthotkey(keystring);
						if (*keystring != ',') break;
						keystring++; // skip the ','
					}
				}
				else if (args.EndOfFile)
				{
					FreeArgs(rda1);
					break;
				}

				FreeArgs(rda1);
			}
			else
			{
				break;
			}
		}

		SelectInput(OldFile);
		Close(file);
	}

	if (numkeys)
	{
		SetIPrefs(actions,(sizeof (struct IAction)) * numkeys, IP_IACTIONS);
	}

	FreeMem(actions,alen);
}

